from functools import reduce
from operator import xor

from two_base.binary import BinaryError, BaseBits, dif_binary_types


class BaseBin:
    """The base of both bin types"""
    def __init__(self, iterable, binary=False):
        for i in iterable:
            if not isinstance(i, BaseBits):
                msg = '{!s} data structure can only contain FrozenBits objects'
                raise BinaryError(msg.format(type(self).__name__))
        self._data = set(iterable)
        self._binary = binary
        self.decode()
        if self._binary:
            self.encode()

    def __repr__(self):
        if self._data:
            return_val = '%s' % self._data
            return '%s(%s)' % (type(self).__name__,
                               return_val.removeprefix('{').removesuffix('}'))
        return '{0.__name__}'.format(type(self))

    def __str__(self):
        return str(tuple(self._data))

    def __format__(self, fmt_spec=''):
        if fmt_spec.endswith('d'):
            fmt_spec = fmt_spec[:-1]
            if self._binary:
                self.decode()
                arg_str = ', '.join([format(i, fmt_spec) for i in self])
                self.encode()
                return '%s(%s)' % (type(self).__name__, arg_str)
            arg_str = ', '.join([format(i, fmt_spec) for i in self])
            return '%s(%s)' % (type(self).__name__, arg_str)
        elif fmt_spec.endswith('c'):
            fmt_spec = fmt_spec[:-1]
            if not self._binary:
                self.encode()
                arg_str = ', '.join([format(i, fmt_spec) for i in self])
                self.decode()
                return '%s(%s)' % (type(self).__name__, arg_str)
            arg_str = ', '.join([format(i, fmt_spec) for i in self])
            return '%s(%s)' % (type(self).__name__, arg_str)
        elif fmt_spec.endswith('o'):
            fmt_spec = fmt_spec[:-1]
            if self._binary:
                self.decode()
                arg_str = ', '.join([format(i, fmt_spec) for i in self])
                self.encode()
                return '%s(%s)' % (type(self).__name__, arg_str)
            else:
                self.encode()
                arg_str = ', '.join([format(i, fmt_spec) for i in self])
                self.decode()
                return '%s(%s)' % (type(self).__name__, arg_str)
        arg_str = ', '.join([format(i, fmt_spec) for i in self])
        return '%s(%s)' % (type(self).__name__, arg_str)

    def __iter__(self):
        return (i for i in self._data)

    def __len__(self):
        return len(self._data)

    def __contains__(self, item):
        try:
            if dif_binary_types(self, item):
                item.encode()
                if item._binary:
                    item.decode()
            if item in self._data:
                return True
            return False
        except (BinaryError, AttributeError):
            return False

    def __or__(self, other):
        if isinstance(other, BaseBin):
            return type(self)(self._data | other._data)
        return NotImplemented

    def __ior__(self, other):
        for i in other:
            self.add(i)
        return self

    def __and__(self, other):
        if isinstance(other, BaseBin):
            return type(self)(self._data & other._data)
        return NotImplemented

    def __iand__(self, other):
        if isinstance(other, BaseBin):
            self._data = self._data & other._data
            return self
        return NotImplemented

    def __sub__(self, other):
        if isinstance(other, BaseBin):
            return type(self)(self._data - other._data)
        return NotImplemented

    def __isub__(self, other):
        if isinstance(other, BaseBin):
            self._data = self._data - other._data
            return self
        return NotImplemented

    def __eq__(self, other):
        return (len(self) == len(other) and
                all(a == b for a, b in zip(self, other)))


class bin(BaseBin):
    """A data structure designated for Binary objects."""
    def add(self, item):
        if not isinstance(item, BaseBits):
            msg = '{!s} data structure can only contain Binary objects'
            raise BinaryError(msg.format(type(self).__name__))
        self._data.add(item)

    def update(self, union):
        self._data.update(union)

    def remove(self, item):
        self._data.remove(item)

    def pop(self):
        return self._data.pop()

    def clear(self):
        self._data.clear()

    def encode(self):
        self._binary = True
        for i in self:
            i.encode()

    def decode(self):
        self._binary = False
        for i in self:
            i.decode()


class frozenbin(BaseBin):
    """Just like 'bin', but it's immutable."""
    def __init__(self, iterable):
        for i in iterable:
            if not isinstance(i, BaseBits):
                msg = '{!s} data structure can only contain FrozenBits objects'
                raise BinaryError(msg.format(type(self).__name__))
        self._data = set(iterable)

    def __ior__(self, other):
        msg = '{!r} is an illegal expression for augmented assignment'
        raise SyntaxError(msg.format(type(self).__name__))

    def __iand__(self, other):
        msg = '{!r} is an illegal expression for augmented assignment'
        raise SyntaxError(msg.format(type(self).__name__))

    def __isub__(self, other):
        msg = '{!r} is an illegal expression for augmented assignment'
        raise SyntaxError(msg.format(type(self).__name__))

    def __eq__(self, other):
        try:
            return (len(self) == len(other) and
                    all(a == b for a, b in zip(self, other)))
        except TypeError:
            return False

    def __hash__(self):
        if self._data:
            hashes = [hash(i) for i in self]
            return reduce(xor, hashes)
        return 0
